@@ -38,18 +38,16 @@ enum EventMeta {
 
 impl EventMeta {
     fn get_event_type(&self) -> Option<&LitStr> {
-        if let Self::Type(lit) = self {
-            Some(lit)
-        } else {
-            None
+        match self {
+            Self::Type(t) => Some(t),
+            _ => None,
         }
     }
 
     fn get_event_kind(&self) -> Option<&EventKind> {
-        if let Self::Kind(lit) = self {
-            Some(lit)
-        } else {
-            None
+        match self {
+            Self::Kind(k) => Some(k),
+            _ => None,
         }
     }
 }
@@ -88,8 +86,8 @@ impl MetaAttrs {
         self.0.iter().find_map(|a| a.get_event_type())
     }
 
-    fn get_event_kinds(&self) -> Vec<&EventKind> {
-        self.0.iter().filter_map(|a| a.get_event_kind()).collect()
+    fn get_event_kind(&self) -> Option<&EventKind> {
+        self.0.iter().find_map(|a| a.get_event_kind())
     }
 }
 
@@ -105,12 +103,6 @@ pub fn expand_event_content(
     input: &DeriveInput,
     ruma_events: &TokenStream,
 ) -> syn::Result<TokenStream> {
-    let ruma_identifiers = quote! { #ruma_events::exports::ruma_identifiers };
-    let serde = quote! { #ruma_events::exports::serde };
-    let serde_json = quote! { #ruma_events::exports::serde_json };
-
-    let ident = &input.ident;
-
     let content_attr = input
         .attrs
         .iter()
@@ -118,22 +110,75 @@ pub fn expand_event_content(
         .map(|attr| attr.parse_args::<MetaAttrs>())
         .collect::<syn::Result<Vec<_>>>()?;
 
-    let event_type = content_attr.iter().find_map(|a| a.get_event_type()).ok_or_else(|| {
-        let msg = "no event type attribute found, \
-            add `#[ruma_event(type = \"any.room.event\", kind = Kind)]` \
-            below the event content derive";
+    let mut event_types: Vec<_> =
+        content_attr.iter().filter_map(|attrs| attrs.get_event_type()).collect();
+    let event_type = match event_types.as_slice() {
+        [] => {
+            return Err(syn::Error::new(
+                Span::call_site(),
+                "no event type attribute found, \
+                 add `#[ruma_event(type = \"any.room.event\", kind = Kind)]` \
+                 below the event content derive",
+            ));
+        }
+        [_] => event_types.pop().unwrap(),
+        _ => {
+            return Err(syn::Error::new(
+                Span::call_site(),
+                "multiple event type attribute found, there can only be one",
+            ));
+        }
+    };
 
-        syn::Error::new(Span::call_site(), msg)
-    })?;
-
-    let content_derives =
-        content_attr.iter().flat_map(|args| args.get_event_kinds()).collect::<Vec<_>>();
+    let mut event_kinds: Vec<_> =
+        content_attr.iter().filter_map(|attrs| attrs.get_event_kind()).collect();
+    let event_kind = match event_kinds.as_slice() {
+        [] => None,
+        [_] => Some(event_kinds.pop().unwrap()),
+        _ => {
+            return Err(syn::Error::new(
+                Span::call_site(),
+                "multiple event kind attribute found, there can only be one",
+            ));
+        }
+    };
 
     // We only generate redacted content structs for state and message events
-    let redacted = if needs_redacted(&content_attr, &content_derives) {
-        let doc = format!("The payload for a redacted `{}`", ident);
-        let redacted_ident = format_ident!("Redacted{}", ident);
-        let kept_redacted_fields = if let syn::Data::Struct(syn::DataStruct {
+    let redacted_event_content = needs_redacted(&content_attr, event_kind)
+        .then(|| generate_redacted_event_content(input, event_type, ruma_events, event_kind))
+        .transpose()?;
+
+    let event_content_impl = generate_event_content_impl(&input.ident, event_type, ruma_events);
+    let static_event_content_impl = event_kind.map(|k| {
+        generate_static_event_content_impl(&input.ident, k, false, event_type, ruma_events)
+    });
+    let marker_trait_impl =
+        event_kind.map(|k| generate_marker_trait_impl(k, &input.ident, ruma_events)).transpose()?;
+
+    Ok(quote! {
+        #redacted_event_content
+        #event_content_impl
+        #static_event_content_impl
+        #marker_trait_impl
+    })
+}
+
+fn generate_redacted_event_content(
+    input: &DeriveInput,
+    event_type: &LitStr,
+    ruma_events: &TokenStream,
+    event_kind: Option<&EventKind>,
+) -> Result<TokenStream, syn::Error> {
+    let ruma_identifiers = quote! { #ruma_events::exports::ruma_identifiers };
+    let serde = quote! { #ruma_events::exports::serde };
+    let serde_json = quote! { #ruma_events::exports::serde_json };
+
+    let ident = &input.ident;
+    let doc = format!("The payload for a redacted `{}`", ident);
+    let redacted_ident = format_ident!("Redacted{}", ident);
+
+    let kept_redacted_fields =
+        if let syn::Data::Struct(syn::DataStruct {
             fields: syn::Fields::Named(syn::FieldsNamed { named, .. }),
             ..
         }) = &input.data
@@ -171,159 +216,133 @@ pub fn expand_event_content(
         } else {
             vec![]
         };
-        let redaction_struct_fields = kept_redacted_fields.iter().flat_map(|f| &f.ident);
 
-        // redacted_fields allows one to declare an empty redacted event without braces,
-        // otherwise `RedactedWhateverEventContent {}` is needed.
-        // The redacted_return is used in `EventContent::redacted` which only returns
-        // zero sized types (unit structs).
-        let (redacted_fields, redacted_return) = if kept_redacted_fields.is_empty() {
-            (quote! { ; }, quote! { Ok(#redacted_ident {}) })
-        } else {
-            (
-                quote! {
-                    { #( #kept_redacted_fields, )* }
-                },
-                quote! {
-                    Err(#serde::de::Error::custom(
-                        format!("this redacted event has fields that cannot be constructed")
-                    ))
-                },
-            )
-        };
+    let redaction_struct_fields = kept_redacted_fields.iter().flat_map(|f| &f.ident);
 
-        let has_deserialize_fields = if kept_redacted_fields.is_empty() {
-            quote! { #ruma_events::HasDeserializeFields::False }
-        } else {
-            quote! { #ruma_events::HasDeserializeFields::True }
-        };
-
-        let has_serialize_fields = if kept_redacted_fields.is_empty() {
-            quote! { false }
-        } else {
-            quote! { true }
-        };
-
-        let initializer = if kept_redacted_fields.is_empty() {
-            let doc = format!("Creates an empty {}.", redacted_ident);
-            quote! {
-                impl #redacted_ident {
-                    #[doc = #doc]
-                    pub fn new() -> Self {
-                        Self
-                    }
-                }
-            }
-        } else {
-            TokenStream::new()
-        };
-
-        let redacted_event_content =
-            generate_event_content_impl(&redacted_ident, event_type, ruma_events);
-
-        let redacted_event_content_derive = content_derives
-            .iter()
-            .map(|kind| match kind {
-                EventKind::Message => quote! {
-                    #[automatically_derived]
-                    impl #ruma_events::RedactedMessageEventContent for #redacted_ident {}
-                },
-                EventKind::State => quote! {
-                    #[automatically_derived]
-                    impl #ruma_events::RedactedStateEventContent for #redacted_ident {}
-                },
-                _ => TokenStream::new(),
-            })
-            .collect::<TokenStream>();
-
-        quote! {
-            // this is the non redacted event content's impl
-            #[automatically_derived]
-            impl #ruma_events::RedactContent for #ident {
-                type Redacted = #redacted_ident;
-
-                fn redact(self, version: &#ruma_identifiers::RoomVersionId) -> #redacted_ident {
-                    #redacted_ident {
-                        #( #redaction_struct_fields: self.#redaction_struct_fields, )*
-                    }
-                }
-            }
-
-            #[doc = #doc]
-            #[derive(Clone, Debug, #serde::Deserialize, #serde::Serialize)]
-            #[cfg_attr(not(feature = "unstable-exhaustive-types"), non_exhaustive)]
-            pub struct #redacted_ident #redacted_fields
-
-            #initializer
-
-            #redacted_event_content
-
-            #[automatically_derived]
-            impl #ruma_events::RedactedEventContent for #redacted_ident {
-                fn empty(ev_type: &str) -> #serde_json::Result<Self> {
-                    if ev_type != #event_type {
-                        return Err(#serde::de::Error::custom(
-                            format!("expected event type `{}`, found `{}`", #event_type, ev_type)
-                        ));
-                    }
-
-                    #redacted_return
-                }
-
-                fn has_serialize_fields(&self) -> bool {
-                    #has_serialize_fields
-                }
-
-                fn has_deserialize_fields() -> #ruma_events::HasDeserializeFields {
-                    #has_deserialize_fields
-                }
-            }
-
-            #redacted_event_content_derive
-        }
+    let (redacted_fields, redacted_return) = if kept_redacted_fields.is_empty() {
+        (quote! { ; }, quote! { Ok(#redacted_ident {}) })
     } else {
-        TokenStream::new()
+        (
+            quote! {
+                { #( #kept_redacted_fields, )* }
+            },
+            quote! {
+                Err(#serde::de::Error::custom(
+                    format!("this redacted event has fields that cannot be constructed")
+                ))
+            },
+        )
     };
 
-    let event_content_derive =
-        generate_event_content_derives(&content_derives, ident, ruma_events)?;
+    let (has_deserialize_fields, has_serialize_fields) = if kept_redacted_fields.is_empty() {
+        (quote! { #ruma_events::HasDeserializeFields::False }, quote! { false })
+    } else {
+        (quote! { #ruma_events::HasDeserializeFields::True }, quote! { true })
+    };
 
-    let event_content = generate_event_content_impl(ident, event_type, ruma_events);
+    let constructor = kept_redacted_fields.is_empty().then(|| {
+        let doc = format!("Creates an empty {}.", redacted_ident);
+        quote! {
+            impl #redacted_ident {
+                #[doc = #doc]
+                pub fn new() -> Self {
+                    Self
+                }
+            }
+        }
+    });
+
+    let redacted_event_content =
+        generate_event_content_impl(&redacted_ident, event_type, ruma_events);
+
+    let marker_trait_impl = match event_kind {
+        Some(EventKind::Message) => quote! {
+            #[automatically_derived]
+            impl #ruma_events::RedactedMessageEventContent for #redacted_ident {}
+        },
+        Some(EventKind::State) => quote! {
+            #[automatically_derived]
+            impl #ruma_events::RedactedStateEventContent for #redacted_ident {}
+        },
+        _ => TokenStream::new(),
+    };
+
+    let static_event_content_impl = event_kind.map(|k| {
+        generate_static_event_content_impl(&redacted_ident, k, true, event_type, ruma_events)
+    });
 
     Ok(quote! {
-        #event_content
-        #event_content_derive
-        #redacted
+        // this is the non redacted event content's impl
+        #[automatically_derived]
+        impl #ruma_events::RedactContent for #ident {
+            type Redacted = #redacted_ident;
+
+            fn redact(self, version: &#ruma_identifiers::RoomVersionId) -> #redacted_ident {
+                #redacted_ident {
+                    #( #redaction_struct_fields: self.#redaction_struct_fields, )*
+                }
+            }
+        }
+
+        #[doc = #doc]
+        #[derive(Clone, Debug, #serde::Deserialize, #serde::Serialize)]
+        #[cfg_attr(not(feature = "unstable-exhaustive-types"), non_exhaustive)]
+        pub struct #redacted_ident #redacted_fields
+
+        #constructor
+
+        #redacted_event_content
+
+        #[automatically_derived]
+        impl #ruma_events::RedactedEventContent for #redacted_ident {
+            fn empty(ev_type: &str) -> #serde_json::Result<Self> {
+                if ev_type != #event_type {
+                    return Err(#serde::de::Error::custom(
+                        format!("expected event type `{}`, found `{}`", #event_type, ev_type)
+                    ));
+                }
+
+                #redacted_return
+            }
+
+            fn has_serialize_fields(&self) -> bool {
+                #has_serialize_fields
+            }
+
+            fn has_deserialize_fields() -> #ruma_events::HasDeserializeFields {
+                #has_deserialize_fields
+            }
+        }
+
+        #static_event_content_impl
+        #marker_trait_impl
     })
 }
 
-fn generate_event_content_derives(
-    content_attr: &[&EventKind],
+fn generate_marker_trait_impl(
+    event_kind: &EventKind,
     ident: &Ident,
     ruma_events: &TokenStream,
 ) -> syn::Result<TokenStream> {
-    let msg = "valid event kinds are GlobalAccountData, RoomAccountData, \
-        EphemeralRoom, Message, State, ToDevice";
-    let marker_traits: Vec<_> = content_attr
-        .iter()
-        .map(|kind| match kind {
-            EventKind::GlobalAccountData => Ok(quote! { GlobalAccountDataEventContent }),
-            EventKind::RoomAccountData => Ok(quote! { RoomAccountDataEventContent }),
-            EventKind::Ephemeral => Ok(quote! { EphemeralRoomEventContent }),
-            EventKind::Message => Ok(quote! { MessageEventContent }),
-            EventKind::State => Ok(quote! { StateEventContent }),
-            EventKind::ToDevice => Ok(quote! { ToDeviceEventContent }),
-            EventKind::Redaction | EventKind::Presence | EventKind::Decrypted => {
-                Err(syn::Error::new_spanned(ident, msg))
-            }
-        })
-        .collect::<syn::Result<_>>()?;
+    let marker_trait = match event_kind {
+        EventKind::GlobalAccountData => quote! { GlobalAccountDataEventContent },
+        EventKind::RoomAccountData => quote! { RoomAccountDataEventContent },
+        EventKind::Ephemeral => quote! { EphemeralRoomEventContent },
+        EventKind::Message => quote! { MessageEventContent },
+        EventKind::State => quote! { StateEventContent },
+        EventKind::ToDevice => quote! { ToDeviceEventContent },
+        EventKind::Redaction | EventKind::Presence | EventKind::Decrypted => {
+            return Err(syn::Error::new_spanned(
+                ident,
+                "valid event kinds are GlobalAccountData, RoomAccountData, \
+                 EphemeralRoom, Message, State, ToDevice",
+            ));
+        }
+    };
 
     Ok(quote! {
-        #(
-            #[automatically_derived]
-            impl #ruma_events::#marker_traits for #ident {}
-        )*
+        #[automatically_derived]
+        impl #ruma_events::#marker_trait for #ident {}
     })
 }
 
@@ -358,10 +377,37 @@ fn generate_event_content_impl(
     }
 }
 
-fn needs_redacted(input: &[MetaAttrs], content_derives: &[&EventKind]) -> bool {
+fn generate_static_event_content_impl(
+    ident: &Ident,
+    event_kind: &EventKind,
+    redacted: bool,
+    event_type: &LitStr,
+    ruma_events: &TokenStream,
+) -> TokenStream {
+    let event_kind = match event_kind {
+        EventKind::GlobalAccountData => quote! { GlobalAccountData },
+        EventKind::RoomAccountData => quote! { RoomAccountData },
+        EventKind::Ephemeral => quote! { EphemeralRoomData },
+        EventKind::Message => quote! { Message { redacted: #redacted } },
+        EventKind::State => quote! { State { redacted: #redacted } },
+        EventKind::ToDevice => quote! { ToDevice },
+        EventKind::Redaction | EventKind::Presence | EventKind::Decrypted => {
+            unreachable!("not a valid event content kind")
+        }
+    };
+
+    quote! {
+        impl #ruma_events::StaticEventContent for #ident {
+            const KIND: #ruma_events::EventKind = #ruma_events::EventKind::#event_kind;
+            const TYPE: &'static ::std::primitive::str = #event_type;
+        }
+    }
+}
+
+fn needs_redacted(input: &[MetaAttrs], event_kind: Option<&EventKind>) -> bool {
     // `is_custom` means that the content struct does not need a generated
     // redacted struct also. If no `custom_redacted` attrs are found the content
     // needs a redacted struct generated.
     !input.iter().any(|a| a.is_custom())
-        && content_derives.iter().any(|e| e.is_message() || e.is_state())
+        && matches!(event_kind, Some(EventKind::Message) | Some(EventKind::State))
 }
